@@ -48,11 +48,13 @@ impl<'a> fmt::Debug for ChangeKind<'a> {
 
 /// Fields that are common to both `Syntax::List` and `Syntax::Atom`.
 pub struct SyntaxInfo<'a> {
+    /// The previous node with the same parent as this one.
+    previous_sibling: Cell<Option<&'a Syntax<'a>>>,
     /// The next node with the same parent as this one.
-    pub next_sibling: Cell<Option<&'a Syntax<'a>>>,
+    next_sibling: Cell<Option<&'a Syntax<'a>>>,
     /// The syntax node that occurs before this one, in a depth-first
     /// tree traversal.
-    pub prev: Cell<Option<&'a Syntax<'a>>>,
+    prev: Cell<Option<&'a Syntax<'a>>>,
     /// The parent syntax node, if present.
     parent: Cell<Option<&'a Syntax<'a>>>,
     /// Does the previous syntax node occur on the same line as the
@@ -76,6 +78,7 @@ pub struct SyntaxInfo<'a> {
 impl<'a> SyntaxInfo<'a> {
     pub fn new() -> Self {
         Self {
+            previous_sibling: Cell::new(None),
             next_sibling: Cell::new(None),
             prev: Cell::new(None),
             parent: Cell::new(None),
@@ -424,6 +427,7 @@ fn set_content_id(nodes: &[&Syntax], existing: &mut HashMap<ContentKey, u32>) {
 }
 
 pub fn init_next_prev<'a>(roots: &[&'a Syntax<'a>]) {
+    set_prev_sibling(roots);
     set_next_sibling(roots);
     set_prev(roots, None);
     set_prev_is_contiguous(roots);
@@ -446,6 +450,21 @@ fn set_unique_id<'a>(nodes: &[&'a Syntax<'a>], next_id: &mut NonZeroU32) {
     }
 }
 
+fn set_prev_sibling<'a>(nodes: &[&'a Syntax<'a>]) {
+    for (i, node) in nodes.iter().enumerate() {
+        if i == 0 {
+            continue;
+        }
+
+        let sibling = nodes.get(i - 1).copied();
+        node.info().previous_sibling.set(sibling);
+
+        if let List { children, .. } = node {
+            set_prev_sibling(children);
+        }
+    }
+}
+
 fn set_next_sibling<'a>(nodes: &[&'a Syntax<'a>]) {
     for (i, node) in nodes.iter().enumerate() {
         let sibling = nodes.get(i + 1).copied();
@@ -571,16 +590,7 @@ pub enum MatchKind {
 }
 
 impl MatchKind {
-    pub fn first_opposite_span(&self) -> Option<SingleLineSpan> {
-        match self {
-            MatchKind::UnchangedToken { opposite_pos, .. } => opposite_pos.first().copied(),
-            MatchKind::NovelLinePart { opposite_pos, .. } => opposite_pos.first().copied(),
-            MatchKind::Novel { .. } => None,
-            MatchKind::NovelWord { .. } => None,
-        }
-    }
-
-    pub fn is_change(&self) -> bool {
+    pub fn is_novel(&self) -> bool {
         matches!(
             self,
             MatchKind::Novel { .. } | MatchKind::NovelWord { .. } | MatchKind::NovelLinePart { .. }
@@ -693,7 +703,7 @@ impl MatchedPos {
         pos: &[SingleLineSpan],
         is_close: bool,
     ) -> Vec<Self> {
-        let kind = match ck {
+        match ck {
             ReplacedComment(this, opposite) => {
                 let this_content = match this {
                     List { .. } => unreachable!(),
@@ -706,13 +716,13 @@ impl MatchedPos {
                     } => (content, position),
                 };
 
-                return split_comment_words(
+                split_comment_words(
                     this_content,
                     // TODO: handle the whole pos here.
                     pos[0],
                     opposite_content,
                     opposite_pos[0],
-                );
+                )
             }
             Unchanged(opposite) => {
                 let opposite_pos = match opposite {
@@ -730,33 +740,60 @@ impl MatchedPos {
                     Atom { position, .. } => position.clone(),
                 };
 
-                MatchKind::UnchangedToken {
+                let opposite_pos_len = opposite_pos.len();
+                let kind = MatchKind::UnchangedToken {
                     highlight,
                     self_pos: pos.to_vec(),
                     opposite_pos,
+                };
+
+                // Create a MatchedPos for every line that `pos` covers.
+                let mut res = vec![];
+                for (i, line_pos) in pos.iter().enumerate() {
+                    // Don't create a MatchedPos for empty positions
+                    // at the start. We will want empty positions on
+                    // multiline atoms elsewhere, as a multline string
+                    // literal may include empty lines.
+                    if i == 0 && line_pos.start_col == line_pos.end_col {
+                        continue;
+                    }
+
+                    res.push(Self {
+                        kind: kind.clone(),
+                        pos: *line_pos,
+                    });
+
+                    // Ensure we have the same number of unchanged
+                    // MatchedPos on the LHS and RHS. This allows us
+                    // to consider unchanged MatchedPos values
+                    // pairwise.
+                    if res.len() == opposite_pos_len {
+                        break;
+                    }
                 }
+                res
             }
-            Novel => MatchKind::Novel { highlight },
-        };
+            Novel => {
+                let kind = MatchKind::Novel { highlight };
+                // Create a MatchedPos for every line that `pos` covers.
+                let mut res = vec![];
+                for line_pos in pos {
+                    // Don't create a MatchedPos for empty positions. This
+                    // occurs when we have lists with empty open/close
+                    // delimiter positions, such as the top-level list of syntax items.
+                    if pos.len() == 1 && line_pos.start_col == line_pos.end_col {
+                        continue;
+                    }
 
-        // Create a MatchedPos for every line that `pos` covers.
-        // TODO: what about opposite pos?
-        let mut res = vec![];
-        for line_pos in pos {
-            // Don't create a MatchedPos for empty positions. This
-            // occurs when we have lists with empty open/close
-            // delimiter positions, such as the top-level list of syntax items.
-            if line_pos.start_col == line_pos.end_col {
-                continue;
+                    res.push(Self {
+                        kind: kind.clone(),
+                        pos: *line_pos,
+                    });
+                }
+
+                res
             }
-
-            res.push(Self {
-                kind: kind.clone(),
-                pos: *line_pos,
-            });
         }
-
-        res
     }
 }
 
@@ -814,14 +851,13 @@ pub fn zip_pad_shorter<Tx: Clone, Ty: Clone>(
 ) -> Vec<(Option<Tx>, Option<Ty>)> {
     let mut res = vec![];
 
-    let mut i = 0;
+    let mut lhs_iter = lhs.iter();
+    let mut rhs_iter = rhs.iter();
     loop {
-        match (lhs.get(i), rhs.get(i)) {
+        match (lhs_iter.next(), rhs_iter.next()) {
             (None, None) => break,
             (x, y) => res.push((x.cloned(), y.cloned())),
         }
-
-        i += 1;
     }
 
     res
@@ -840,18 +876,16 @@ pub fn zip_repeat_shorter<Tx: Clone, Ty: Clone>(lhs: &[Tx], rhs: &[Ty]) -> Vec<(
     };
 
     let mut res = vec![];
-
-    let mut i = 0;
+    let mut lhs_iter = lhs.iter();
+    let mut rhs_iter = rhs.iter();
     loop {
-        match (lhs.get(i), rhs.get(i)) {
+        match (lhs_iter.next(), rhs_iter.next()) {
             (None, None) => break,
             (x, y) => res.push((
                 x.cloned().unwrap_or_else(|| lhs_last.clone()),
                 y.cloned().unwrap_or_else(|| rhs_last.clone()),
             )),
         }
-
-        i += 1;
     }
 
     res
